length_units = {
    "English2": {
        "SUL": "inch",
        "MUL": "feet",
        "LUL": "mile",
        "PerLUL": "ohm/mile",
    },
    "English1": {
        "SUL": "inch",
        "MUL": "feet",
        "LUL": "kft",
        "PerLUL": "ohm/kft",
    },
    "English": {
        "SUL": "inch",
        "MUL": "feet",
        "LUL": "kft",
        "PerLUL": "ohm/kft",
    },
    "Metric": {
        "SUL": "mm",
        "MUL": "m",
        "LUL": "km",
        "PerLUL": "ohm/km",
    },
}
